@@ -22,6 +22,9 @@ def ponderomotive_energy(E, material, laser):
 def get_critical_energy(E, material, laser):
     return (1+material.m_red/material.m_VB) * (material.bandgap + ponderomotive_energy(E, material, laser))
 
+def critical_plasma_density(material, laser):
+    return c.epsilon_0*material.m_red*c.m_e/c.e**2*(laser.omega**2 + material.damping**2)
+
 def ee_coll_freq(Ekin, material):
     """ electron-electron collision rate """
     return 4*c.pi*c.epsilon_0/c.e**2*(6/material.m_CB/c.m_e)**0.5*(2*Ekin/3)**1.5
@@ -50,7 +53,7 @@ def el_Ekin_max(E, material, laser):
         /(2.*Ec*material.cross_section*material.density)*(material.m_CB*c.m_e*c.pi/(3.*Ec))**.5))).max()
 
 
-def surface_roughness(material, boundary, amplitude, noise='white', feature_size='default', show=False):
+def surface_roughness(material, boundary, amplitude=1, noise='white', feature_size='default', fill_factor=0.1, show=False, roughness_map=None):
 
     if not hasattr(material, 'domain'):
         raise ValueError('Material has to be attached to a Domain in order to add surface roughness to it.')
@@ -58,42 +61,55 @@ def surface_roughness(material, boundary, amplitude, noise='white', feature_size
     if boundary != 'xmin':
         raise NotImplementedError('Surface roughness can only be attached to the xmin boundary.')
 
-    if noise.lower() == 'white':
-        roughness_map = white_roughness_map(shape=(material.domain.Ny,material.domain.Nz))
+    if roughness_map is None:
+        if noise.lower() == 'white':
+            roughness_map = white_roughness_map(shape=(material.domain.Ny,material.domain.Nz))
 
-    else:
-        feature_size = amplitude if feature_size is 'default' else feature_size
-        res = (int(material.domain.Ly/feature_size),int(material.domain.Lz/feature_size))
-        if noise.lower() == 'perlin':
-            roughness_map = perlin_roughness_map(shape=(material.domain.Ny,material.domain.Nz), res=res)
-        if noise.lower() == 'fractal':
-            roughness_map = fractal_roughness_map(shape=(material.domain.Ny,material.domain.Nz), res=res)
+        if noise.lower() == 'binary':
+            roughness_map = binary_roughness_map(shape=(material.domain.Ny,material.domain.Nz), fill_factor=fill_factor)
 
-    roughness_map *= amplitude
+        if noise.lower() in ['perlin', 'fractal']:
+            feature_size = amplitude if feature_size == 'default' else feature_size
+            res = (int(material.domain.Ly/feature_size),int(material.domain.Lz/feature_size))
+            if noise.lower() == 'perlin':
+                roughness_map = perlin_roughness_map(shape=(material.domain.Ny,material.domain.Nz), res=res)
+            if noise.lower() == 'fractal':
+                roughness_map = fractal_roughness_map(shape=(material.domain.Ny,material.domain.Nz), res=res)
+
+        roughness_map *= amplitude
 
     if show:
         fig = plt.figure()
         plt.imshow(roughness_map, cmap='gray')
         fig.canvas.set_window_title('Roughness map')
         plt.show(block=False)
+        plt.pause(0.01)
 
     ix, rem = material.parse_index(material.boundaries[boundary]/material.domain.dx)
 
     for iy in range(0,material.domain.Ny-1):
         for iz in range(0,material.domain.Nz-1):
             ir, rem_r = material.parse_index((material.boundaries[boundary]-roughness_map[iy,iz])/material.domain.dx)
-            material.mask[ir:ix+1,iy,iz] = 1
-            material.mask[ir,iy,iz] -= rem_r
+            if roughness_map[iy,iz] >= 0:
+                material.mask[ir:ix+1,iy,iz] = 1
+                material.mask[ir,iy,iz] -= rem_r
+            else:
+                material.mask[ir:ix+1,iy,iz] = 0
+                material.mask[ir,iy,iz] += rem_r
 
     for iy in range(0,material.domain.Ny):
         material.mask[:,iy,-1] = material.mask[:,iy,0]
     for iz in range(0,material.domain.Nz):
         material.mask[:,-1,iz] = material.mask[:,0,iz]
 
+    return roughness_map
+
 
 def white_roughness_map(shape):
     return np.random.random(shape)
 
+def binary_roughness_map(shape, fill_factor):
+    return np.random.choice([0.,1.], shape, p=[1-fill_factor, fill_factor])
 
 def perlin_roughness_map(shape, res):
     def f(t):
