@@ -1,16 +1,14 @@
 import numpy as np
 import copy
-import matplotlib.pyplot as plt
-import matplotlib.patches as patches
 import gzip
 import pickle
 
 from .misc import *
+from .viewer import *
 from .backend import backend as bd
 
 
 
-
 class Observer():
 
     def __init__(self, target:str=None, mode:str='watch', x=None, y=None, z=None, keep_pml=False, out_step=1):
@@ -69,8 +67,9 @@ class Observer():
         #### Plasma density, energies and formation rates ##################################
         # TODO : get the actual list of material's tracktables
         if target in ['rho','rho_k','rho_hk','Ekin','Ekin_h',
-                      'el_heating_rate','critical_energy',
-                      'r_e','r_h','xi_e','xi_h','coll_freq_en','coll_freq_hn']:
+                      'el_heating_rate','hl_heating_rate','critical_energy',
+                      'r_e','r_h','xi_e','xi_h','coll_freq_en','coll_freq_hn',
+                      'trapped_rho']:
             data = 0
             for material in self.domain.materials:
                 data += getattr(material, target)
@@ -111,7 +110,7 @@ class Observer():
             if self.x is not None:
                 data = data[int((self.x-self.domain.x.min())/self.domain.dx)]
 
-            if not self.keep_pml and self.x is None and self.domain.D > 0:
+            if not self.keep_pml and self.x is None and self.domain.D > 0 and self.domain.nb_pml > 0:
                 data = data[self.domain.nb_pml:-self.domain.nb_pml]
 
         if data is not None:
@@ -119,6 +118,18 @@ class Observer():
         else:
             raise ValueError(f'Could not get {target} from simulation domain.')
 
+    
+    def get_fourier_data(self, target=None, slicing=True):
+        data = self.get_data(target, slicing)
+
+        fft = np.fft.fft if self.D < 2 else np.fft.fft2
+
+        complex_fourier_data = fft(data)
+        fourier_data = (complex_fourier_data.real**2 + complex_fourier_data.imag**2)**0.5
+        fourier_data = np.fft.fftshift(fourier_data)
+
+        return fourier_data
+
 
     def terminate(self):
         pass
@@ -136,9 +147,12 @@ class Watcher(Observer):
     # TODO: Documentation
     # TODO: Log graphs
 
-    def __init__(self, target:str=None, x=None, y=None, z=None, vlim=(0,0), figsize='default', c='C0', colormap='seismic', keep_pml=False, loop=False, out_step=1):
+    def __init__(self, target:str=None, x=None, y=None, z=None, fourier:bool=False, 
+                 vlim=(0,0), figsize='default', c='C0', colormap='seismic', keep_pml=False, 
+                 loop=False, out_step=1):
         super(Watcher, self).__init__(target, 'watch', x, y, z, keep_pml, out_step)
 
+        self.fourier = fourier
         self.vlim = vlim
         self.figsize = figsize
         self.c = c
@@ -152,7 +166,8 @@ class Watcher(Observer):
 
     def call(self, data=None):
 
-        data = self.get_data() if data is None else data
+        get_data = self.get_data if not self.fourier else self.get_fourier_data
+        data = get_data() if data is None else data
 
         if self.save_data:
             self.stack_data.append(copy.deepcopy(data))
@@ -172,7 +187,7 @@ class Watcher(Observer):
                     axis = {'y': np.linspace(0, self.domain.Ly, self.domain.Ny)}
                 if self.z is None:
                     axis = {'z': np.linspace(0, self.domain.Lz, self.domain.Nz)}
-                self.viewer = Viewer1d(axis, self.domain, self.target, self.vlim, self.c, self.keep_pml, self.figsize)
+                self.viewer = Viewer1d(axis, self.domain, self.target, self.fourier, self.vlim, self.c, self.keep_pml, self.figsize)
 
             if self.D == 2:
                 if self.x is not None:
@@ -184,7 +199,7 @@ class Watcher(Observer):
                 if self.z is not None:
                     axes = {'x': np.linspace(0, self.domain.Lx, self.domain.Nx),
                             'y': np.linspace(0, self.domain.Ly, self.domain.Ny)}
-                self.viewer = Viewer2d(axes, self.domain, self.target, self.vlim, self.colormap, self.keep_pml, self.figsize)
+                self.viewer = Viewer2d(axes, self.domain, self.target, self.fourier, self.vlim, self.colormap, self.keep_pml, self.figsize)
 
             if self.D == 3:
                 raise NotImplementedError('3D view is not available.')
@@ -195,13 +210,19 @@ class Watcher(Observer):
 
         
 
-
     def terminate(self):
 
         # FIXME: loop over multiple windows doesn't work because stuck in while loop
 
         self.save_data = False
 
+        if self.loop:
+            print(f'Looping over {self.target}.')
+            self.start_loop()
+
+
+    def start_loop(self):
+
         while self.loop:
 
             if self.D == 0:
@@ -214,6 +235,7 @@ class Watcher(Observer):
             if type(self.loop) == int:
                 self.loop -= 1
 
+
     def exit_loop(self, event):
         self.loop = False
         
@@ -228,16 +250,15 @@ class Dumper(Observer):
 
         self.zipfile = gzip.GzipFile(f'{target}.zip', 'wb')
 
-
     def call(self):
         data = self.get_data()
         self.zipfile.write(pickle.dumps(data))
 
-
     def terminate(self):
         self.zipfile.close()
 
 
+
 class Printer(Observer):
 
     # TODO: Documentation
@@ -245,18 +266,17 @@ class Printer(Observer):
     def __init__(self, target:str=None, x=None, y=None, z=None, keep_pml=False, out_step=1):
         super(Printer, self).__init__(target, 'print', x, y, z, keep_pml, out_step)
 
-
     def call(self):
         print(self.get_data().mean())
 
 
+
 class Returner(Observer):
 
     # TODO: Documentation
 
     def __init__(self, target:str=None, x=None, y=None, z=None, keep_pml=False, out_step=1):
         super(Returner, self).__init__(target, 'return', x, y, z, keep_pml, out_step)
-
         self.stack_data = []
 
     def call(self, data=None):
@@ -268,184 +288,16 @@ class Returner(Observer):
 
 
 
+class Stopper(Observer):
 
+    def __init__(self, target:str, condition:callable, x=None, y=None, z=None, verbose:bool=True):
+        super(Stopper, self).__init__(target, 'stop', x, y, z, keep_pml=False, out_step=1)
+        self.condition = condition
+        self.verbose = verbose
 
-class Viewer():
-
-    def __init__(self, domain, target='', vlim=(0,0), figsize='default'):
-        self.domain = domain
-        self.target = target
-
-        self.min_value = vlim[0]
-        self.max_value = vlim[1]
-
-        self.fig = plt.figure() if figsize == 'default' else plt.figure(figsize=figsize)
-        self.fig.canvas.set_window_title(self.target)
-        self.fig.show()
-        self.fig.canvas.draw()
-        self.fig.subplots_adjust(bottom=0.2)
-
-    def update(self, time, data):
-
-        self.min_value = min(self.min_value, data.min())
-        self.max_value = max(self.max_value, data.max())
-
-
-
-class Viewer0d(Viewer):
-
-    def __init__(self, domain, target, vlim=(0,0), c='C0', figsize='default'):
-        super(Viewer0d, self).__init__(domain, target, vlim, figsize)
-
-        self.times = []
-        self.data_axis = []
-
-        self.time_axis, self.units_factor, units_name = format_axis(domain.times, mode='time')
-
-        self.ax = self.fig.add_subplot(111)
-        self.ax.set_xlim(min(self.time_axis),max(self.time_axis))
-        self.line = self.ax.plot(self.times, self.data_axis, c=c)[0]
-        self.ax.set_xlabel(f't [{units_name}]')
-
-
-
-    def update(self, time, data):
-        super(Viewer0d, self).update(time, data)
-
-        self.times.append(time/self.units_factor)
-        self.data_axis.append(data)
-        self.line.set_data(self.times, self.data_axis)
-        self.ax.set_ylim(self.min_value, self.max_value)
-
-        self.fig.canvas.draw()
-        self.fig.canvas.flush_events()
-
-
-
-
-class Viewer1d(Viewer):
-
-    def __init__(self, axis, domain, target, vlim=(0,0), c='C0', show_pml=False, figsize='default'):
-        super(Viewer1d, self).__init__(domain, target, vlim, figsize)
-
-        axis_name = list(axis.keys())[0]
-        axis_values, units_factor, units_name = format_axis(axis[axis_name], mode='length')
-
-        self.ax = self.fig.add_subplot(111)
-        self.ax.set_xlim(axis_values.min(), axis_values.max())
-        self.line = self.ax.plot(axis_values, np.zeros(axis_values.shape[0]), c=c)[0]
-
-        for material in self.domain.materials:
-            self.ax.add_patch(patches.Rectangle((material.boundaries[f'{axis_name}min']/units_factor,-1e90),
-                                                (material.boundaries[f'{axis_name}max']-material.boundaries[f'{axis_name}min'])/units_factor,
-                                                 1e200, linewidth=1.5, edgecolor='0.8', facecolor='0.9'))
-        
-        if axis_name == 'x':
-            if show_pml:
-                self.ax.add_patch(patches.Rectangle((self.domain.x.min(), -1e90),
-                                                    domain.pml_width/units_factor, 1e200,
-                                                    linewidth=1.5, edgecolor='0.6', facecolor='0.7'))
-                self.ax.add_patch(patches.Rectangle(((domain.Lx-domain.pml_width)/units_factor, -1e90),
-                                                    domain.pml_width/units_factor, 1e200,
-                                                    linewidth=1.5, edgecolor='0.6', facecolor='0.7'))
-            else:
-                self.ax.set_xlim(domain.pml_width/units_factor, (domain.Lx-domain.pml_width)/units_factor)
-
-        self.ax.axvline(self.domain.laser.position, color='r', lw=2)
-
-        self.ax.set_xlabel(f'{axis_name} [{units_name}]')
-
-
-    def update(self, time, data):
-        super(Viewer1d, self).update(time, data)
-
-        self.line.set_ydata(data)
-        self.ax.set_ylim(self.min_value, self.max_value)
-
-        self.fig.canvas.draw()
-        self.fig.canvas.flush_events()
-
-
-
-class Viewer2d(Viewer):
-
-    def __init__(self, axes, domain, target, vlim=(0,0), colormap='seismic', show_pml=False, figsize='default'):
-        super(Viewer2d, self).__init__(domain, target, vlim, figsize)
-
-        self.colormap = colormap
-        self.axes = axes
-        self.show_pml = show_pml
-        self.ax = self.fig.add_subplot(111)
-        self.img = None
-
-    def update(self, time, data):
-        # super(Viewer2d, self).update(time, data)
-        # TODO: Optional rotation
-
-        # data = np.rot90(data)
-
-        if self.img is None:
-
-            extent = []
-
-            for axis in ['z','y','x']:
-
-                if axis not in self.axes:
-                    continue
-
-                if axis == 'x' and not self.show_pml:
-                    no_pml_ax = self.axes[axis][self.domain.nb_pml:-self.domain.nb_pml]
-                else:
-                    no_pml_ax = self.axes[axis]
-
-                if len(extent) == 0:
-                    axis_values, units_factor, units_name = format_axis(no_pml_ax, mode='length')
-                    self.ax.set_xlabel(f'{axis} [{units_name}]')
-                else:
-                    axis_values = no_pml_ax/units_factor
-                    self.ax.set_ylabel(f'{axis} [{units_name}]')
-
-                extent.append(axis_values.min())
-                extent.append(axis_values.max())
-
-            self.img = self.ax.imshow(data, extent=extent, cmap=self.colormap, aspect=1)
-
-        self.img.set_data(data)
-        self.img.set_clim(data.min(), data.max())
-
-        self.fig.canvas.draw()
-        self.fig.canvas.flush_events()
-
-
-
-
-def format_axis(ax, mode='length'):
-
-    # TODO: use the format value function in misc.py
-
-    axis = copy.deepcopy(ax)
-    max_value = max([abs(axis.max()), abs(axis.min())])
-    factor, units = 1, ''
-
-    if max_value < 1 and max_value >= 1e-3:
-        axis *= 1e3
-        factor, units = 1e-3, 'm'
-    if max_value < 1e-3 and max_value >= 1e-6:
-        axis *= 1e6
-        factor, units = 1e-6, r'$\mu$'
-    if max_value < 1e-6 and max_value >= 1e-9:
-        axis *= 1e9
-        factor, units = 1e-9, 'n'
-    if max_value < 1e-9 and max_value >= 1e-12:
-        axis *= 1e12
-        factor, units = 1e-12, 'p'
-    if max_value < 1e-12:
-        axis *= 1e15
-        factor, units = 1e-15, 'f'
-
-    if mode == 'length':
-        units += 'm'
-    if mode == 'time':
-        units += 's'
-
-    return axis, factor, units
\ No newline at end of file
+    def call(self):
+        data = self.get_data()
+        if self.condition(data):
+            self.domain.running = False
+            if self.verbose:
+                print(f'Stopped simulation because the condition on {self.target} is True.')
