@@ -71,7 +71,7 @@ class Domain():
             self.fields[field] = bd.zeros(self.field_shape)
 
 
-    def run(self, time, stability_factor:float=0.95, Nt:int='auto', progress_bar:bool=True):            
+    def run(self, time, stability_factor:float=0.95, Nt:int='auto', progress_bar:bool=True, verbose:bool=True):            
 
         if type(time) == tuple:
             self.start_time = time[0]
@@ -89,7 +89,7 @@ class Domain():
             self.dt = self.Lt/self.Nt
         self.times = np.linspace(self.start_time, self.end_time, self.Nt)
 
-        if self.D > 0:
+        if self.D > 0 and verbose:
             fdx, fdy, fdz = [format_value(d,'m') for d in [self.dx, self.dy, self.dz]]
             print(f'Space discretization: dx={fdx[0]:.1f}{fdx[2]}, dy={fdy[0]:.1f}{fdy[2]}, dz:{fdz[0]:.1f}{fdz[2]}.')
             fdt = format_value(self.dt,'s')
@@ -103,6 +103,8 @@ class Domain():
         else:
             self.tqdm_times = self.times
 
+        self.running = True
+
         for self.it, self.t in enumerate(self.tqdm_times):
 
             if self.D == 0:
@@ -120,6 +122,9 @@ class Domain():
                 self.update_H()
                 self.observe()
 
+            if not self.running:
+                break
+
         return self.terminate()
 
 
@@ -136,12 +141,13 @@ class Domain():
     def update_plasma(self):
         E_amp = self.E_amp
         for material in self.materials:
-            if material.rate_equation is not 'none':
+            if material.rate_equation != 'none':
                 material.field_ionization(E_amp)
             if material.rate_equation in ['sre','mre','dre']:
                 material.impact_ionization(E_amp)
-            if material.rate_equation is not 'none':
+            if material.rate_equation != 'none':
                 material.recombination()
+                material.trapping()
 
 
     def update_currents(self):
@@ -152,7 +158,7 @@ class Domain():
     def update_bounded_current(self):
         # See https://arxiv.org/abs/1603.09410
         for material in self.materials:
-            w0 = 2*c.pi*c.c/material.resonance if material.resonance > 0 else 0
+            w0 = 2*c.pi*c.c/material.resonance if material.resonance > 0 else 1/self.dt
             P = c.epsilon_0*((material.index**2 - 1)*self.fields['E'] + 
                               material.chi2*self.fields['E']**2 + 
                               material.chi3*self.fields['E']**3)
@@ -163,11 +169,15 @@ class Domain():
 
     def update_free_current(self):
         for material in self.materials:
+            
             if material.drude:
                 G = material.damping*self.dt/2
                 self.fields['Jf'] = material.mask[...,None]*(self.fields['Jf']*(1-G)/(1+G) \
                                     + self.dt*c.epsilon_0*material.plasma_freq[...,None]**2*self.fields['E']/(1+G))
 
+            else:
+                self.fields['Jf'] = material.mask[...,None]*2*c.epsilon_0*material.index*material.index_imag*self.laser.omega*self.fields['E']
+
 
 
     def update_E(self):
@@ -189,14 +199,15 @@ class Domain():
             # Section 3.0 of : https://studylib.net/doc/8392930/6.-total-field---scattered-field-fdtd-implementation-in-m...
             # Section 3.10 of : https://www.eecs.wsu.edu/~schneidj/ufdtd/chap3.pdf
             if self.laser is not None:
-                laser_E = self.laser.E(self.t)
 
                 ramp = (np.exp((self.it/10)**2)-1)/np.exp((self.it/10)**2) if self.laser_ramp and self.it < 30 else 1
 
                 if self.laser.source_mode.lower() == 'tfsf':
-                    self.fields['E'][self.laser.index_in_domain,...,2] += ramp*self.dt/(c.epsilon_0*self.dx)*laser_E/(120*c.pi)
+                    self.fields['E'][self.laser.index_in_domain+1,...,2] += ramp*self.laser.E(self.t+self.dt/2)*c.c*self.dt/self.dx
+                elif self.laser.source_mode.lower() == 'soft':
+                    self.fields['E'][self.laser.index_in_domain,...,2] += 2*ramp*self.laser.E(self.t+self.dt/2)*c.c*self.dt/self.dx
                 elif self.laser.source_mode.lower() == 'hard':
-                    self.fields['E'][self.laser.index_in_domain,...,2] = ramp*laser_E
+                    self.fields['E'][self.laser.index_in_domain,...,2] = ramp*self.laser.E(self.t+self.dt/2)
 
             # boundaries
             for boundary in self.boundaries:
@@ -212,15 +223,9 @@ class Domain():
             self.fields['H'] -= self.dt/c.mu_0 * self.curl_E(self.fields['E'])
 
             # add sources
-            if self.laser is not None and self.D > 0:
-                laser_E = self.laser.E(self.t)
-
+            if self.laser is not None and self.D > 0 and self.laser.source_mode.lower() == 'tfsf':
                 ramp = (np.exp((self.it/10)**2)-1)/np.exp((self.it/10)**2) if self.laser_ramp and self.it < 30 else 1
-
-                if self.laser.source_mode.lower() == 'tfsf':
-                    self.fields['H'][self.laser.index_in_domain,...,1] -= ramp*self.dt/(c.mu_0*self.dx)*laser_E
-                elif self.laser.source_mode.lower() == 'hard':
-                    pass
+                self.fields['H'][self.laser.index_in_domain,...,1] -= ramp*self.dt/(c.mu_0*self.dx)*self.laser.E(self.t)
 
             # boundaries
             for boundary in self.boundaries:
@@ -237,7 +242,7 @@ class Domain():
 
     def observe(self):
         for observer in self.observers:
-            if self.it%observer.out_step == 0:
+            if self.it%observer.out_step == 0 and observer.out_step > 0:
                 observer.call()
 
 
@@ -246,6 +251,9 @@ class Domain():
         out_dico = {}
         for observer in self.observers:
 
+            if observer.out_step == -1:
+                observer.call()
+
             if observer.mode == 'return':
                 out_dico[observer.target] = observer.terminate()
             else:
@@ -294,13 +302,15 @@ class Domain():
 
     def set_boundaries(self, dt:float=0):
 
+        self.boundaries = []
+
         if self.D == 0:
-            self.boundaries = []
             return None
 
         if self.D > 0:
             self.nb_pml = int(self.pml_width/self.dx)
-            self.boundaries = [PML(self, 'xmin', dt), PML(self, 'xmax', dt)]
+            if self.nb_pml > 0:
+                self.boundaries += [PML(self, 'xmin', dt), PML(self, 'xmax', dt)]
             self.boundaries += [Periodic(self, 'y'), Periodic(self, 'z')]
         
 
@@ -338,15 +348,3 @@ class Domain():
 
         return curl
 
-
-
-
-# class Time():
-
-#     def __init__(self, start, end, Nt):
-#         self.t = np.linspace(start, end, int(Nt))
-#         self.dt = self.t[1] - self.t[0]
-#         self.Nt = Nt
-
-
-
